/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.BuildOsi7Telegram
 *
 * de.bsvrz.kex.tls.osi7.BuildOsi7Telegram is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.BuildOsi7Telegram is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.BuildOsi7Telegram.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.BuildOsi7Telegram.
 *
 *  de.bsvrz.kex.tls.osi7.BuildOsi7Telegram ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.BuildOsi7Telegram wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */



package de.bsvrz.kex.tls.osi7;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.kex.tls.osi2osi3.osi3.DestinationUnreachableException;
import de.bsvrz.kex.tls.osi2osi3.osi3.NetworkLayer;
import de.bsvrz.kex.tls.osi2osi3.osi3.NetworkLayerListener;
import de.bsvrz.kex.tls.osi2osi3.osi3.TlsNetworkLayer;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

//~ KLASSEN ===================================================================

/**
 * TBD: Kurzbeschreibung. TBD: Langbeschreibung.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class BuildOsi7Telegram extends Thread {

    // ------------------------------------------------------------------------
    // VARIABLENDEKLARATIONEN
    // ------------------------------------------------------------------------
    // Sonstige ---------------------------------------------------------------

    /** DebugLogger für Debug-Ausgaben */
    private static final Debug debug = Debug.getLogger();

    //~ FELDER ================================================================

    private final int COUNT_DEBLOCKS_INDEX = 8;  // Anzahl der De-Blöcke des Einzeltelegramme

    /** Inices für den Aufbau des Telegrammkopfes */
    private final int LENGTH_SINGLETELEGRAM_INDEX = 4;  // Länge des Einzeltelegramms

    /** Map mit Zuordnungen von Blocktransfers zu De. */
    private Map<De, MultiBlockTransfer> _blockTransfers = new HashMap<De, MultiBlockTransfer>();
    private Queue                       _deBlockQueue;

    /** Klasse mit den Definitionen der nicht zusammenzufassenden DE-Bl&ouml;cke. */
    private DeBlocksMultiBlockTransfer _deBlocksMultiBlockTransfer;

    // Private ----------------------------------------------------------------
    private List<DeBlockToSend>   _deBlocksToSend;
    private int                   _jobNumber;
    private NetworkLayer          _networkLayer;
    private NewRouting            _newRouting;
    private SendingOsi7AdressNull _sendingOsi7AdressNull;
    private boolean               _sendOnlyConnectedSM;
    private boolean               _sendOnlyDeOk;

    //~ KONSTRUKTOREN  (und vom Konstruktor verwendete Methoden) ==============

    /**
     * Erzeugt ein Objekt vom Typ BuildOsi7Telegram
	 *
	 * @param deBlockQueue Quelle
	 * @param networkLayer NetworkLayer
	 * @param sendOnlyDeOk sendOnlyDeOk
	 * @param sendOnlyConnectedSM sendOnlyConnectedSM
	 */
    public BuildOsi7Telegram(Queue deBlockQueue, NetworkLayer networkLayer, boolean sendOnlyDeOk, boolean sendOnlyConnectedSM) {
        _deBlockQueue          = deBlockQueue;
        _networkLayer          = networkLayer;
        _sendingOsi7AdressNull = SendingOsi7AdressNull.getInstance();
        _newRouting            = NewRouting.getInstance();
        _sendOnlyDeOk          = sendOnlyDeOk;
        _sendOnlyConnectedSM   = sendOnlyConnectedSM;

        // Instanz der Klasse DeBlocksMultiBlockTransfer holen.
        _deBlocksMultiBlockTransfer = DeBlocksMultiBlockTransfer.getInstance();
    }

    //~ METHODEN ==============================================================

    /**
     * Erzeugt und versendet OSI7-Telegramme f&uuml;r eine Liste von zu versendenden
     * DE-Bl&ouml;cken.
     *
     * @param deBlocks zu versendende DE-Bl&ouml;cke
     * @param sendAsSingleJob sende die OSI7-Telegramme mit gleicher Jobnummer ja/nein
     */
    private void erzeugeOsi7Telegramme(List<DeBlockToSend> deBlocks, boolean sendAsSingleJob) {
        int fg;
        int id;
        int osi7Adress;
        int jobNumber = -1;
        int maxByteCountForDeBlocks;
        int lengthSingleTelegram;
        int countDeBlocks;
        int lengthDeBlock;

        while (!deBlocks.isEmpty()) {
            boolean isTelegramToSend = false;
            TlsData tlsData          = null;

            maxByteCountForDeBlocks = 233 - 4;  // 229 Byte stehen maximal für DE-Blöcke zur Verfügung
            lengthSingleTelegram    = 4;        // Minimale Länge ohne DE-Blöcke
            countDeBlocks           = 0;
            lengthDeBlock           = 0;

            // Ersten DeBlock für dieses Einzeltelegramm lesen
            DeBlockToSend deBlockToSend = deBlocks.remove(0);  // erstem Eintrag holen und aus Liste löschen
            DeBlock       deBlock       = deBlockToSend.getDeBlock();
            De            de            = deBlockToSend.getDe();

            // Prüfen, ob nur DE-Blocke versendet werden sollen, bei denen
            // für die De eine De-Gutmeldung anliegt.
            if (de.isOk() ||!_sendOnlyDeOk) {

                // Telegramme mit mehrere DE-Blöcken und EINEM Einzeltelegramm erstellen.
                // Osi7Telegramm mit MEHREREN Einzeltelegrammen wird NICHT unterstützt,
                // da ohnehin diese Zusammanstellung nur sehr selten möglich ist und
                // zudem viele existente Streckenstationen damit nicht korrekt
                // umgehen können!
                //
                // Osi7Telegramm bauen, nur ein Einzeltelegramm, mehrereDe-Blöcke
                // Länge ist dann 4 (Osi7Header) + 5 (EinzeltelegrammHeader) + Länge aller DeBlöcke
                isTelegramToSend = true;
                tlsData          = new TlsData(9);

                // Knotennummer , FG, ID und Jobnummer ermitteln
                osi7Adress = getOsi7Adress(deBlock);
                fg         = deBlock.getFg();
                id         = deBlock.getId();

                int[] jobNummernBereichFuerLink = new int[0];

                if(_networkLayer instanceof TlsNetworkLayer) {
                    jobNummernBereichFuerLink = ((TlsNetworkLayer) _networkLayer).getJobNummernBereichFuerLink(osi7Adress);
                }

                if(jobNummernBereichFuerLink == null) // Default-Werte verwenden
                {
                    jobNummernBereichFuerLink = new int[2];
                    jobNummernBereichFuerLink[0] = 1;
                    jobNummernBereichFuerLink[1] = 255;
                }

                if(!sendAsSingleJob || jobNumber == -1) {
                    jobNumber = getJobNumber(jobNummernBereichFuerLink[0], jobNummernBereichFuerLink[1]);
                }

                // Allgemeinen Telegrammkopf schreiben
                tlsData.writeIntAsLowHighExtraHighByte(osi7Adress);
                tlsData.writeIntAsByte(1);  // immer genau ein Einzeltelegramm

                // Einzeltelegrammkopf schreiben
                tlsData.writeIntAsByte(lengthSingleTelegram);  // Startlänge ist 4 (noch kein DE-Block)
                tlsData.writeIntAsByte(fg);
                tlsData.writeIntAsByte(id);
                tlsData.writeIntAsByte(jobNumber);
                tlsData.writeIntAsByte(countDeBlocks);         // noch kein De-Block enthalten

                // Ersten De-Block anhängen (noch ohne Längenprüfung, einer geht immer...)
                // Länge Einzeltelegramm und Anzahl De-Blöcke im Einzeltelegrammkopf anpassen
                lengthDeBlock = deBlock.getTlsDeDataBlock().length;
                tlsData.writeByteArray(deBlock.getTlsDeDataBlock());
                lengthSingleTelegram += lengthDeBlock;
                tlsData.writeIntAsByte(lengthSingleTelegram, LENGTH_SINGLETELEGRAM_INDEX);  // Startlänge korrigieren
                tlsData.writeIntAsByte(++countDeBlocks, COUNT_DEBLOCKS_INDEX);              // Anzahl De-Blöcke korrigieren

                // Mögliche Restlänge ermitteln
                maxByteCountForDeBlocks -= lengthDeBlock;

                // Wenn mit gleicher Jobnummer versendet werden soll, keine weiteren DE-Blöcke hinzufügen,
                // das sonst die Reihenfolge zerstört werden kann (kurzer 1. und letzter Block)
                if (!sendAsSingleJob) {

                    // Jetzt versuchen weitere De-Blöcke hinzuzupacken...
                    while ((deBlockToSend = getNextDeBlockToSend(deBlocks, osi7Adress, fg, id, maxByteCountForDeBlocks)) != null) {

                        // Möglichen De-Block mit übereinstimmender Knotennummer, FG, ID und Länge suchen
                        // n-ten De-Block anhängen
                        deBlock       = deBlockToSend.getDeBlock();
                        de            = deBlockToSend.getDe();
                        lengthDeBlock = deBlock.getTlsDeDataBlock().length;
                        tlsData.writeByteArray(deBlock.getTlsDeDataBlock());
                        lengthSingleTelegram += lengthDeBlock;
                        tlsData.writeIntAsByte(lengthSingleTelegram, LENGTH_SINGLETELEGRAM_INDEX);  // Startlänge korrigieren
                        tlsData.writeIntAsByte(++countDeBlocks, COUNT_DEBLOCKS_INDEX);              // Anzahl De-Blöcke korrigieren

                        // Mögliche Restlänge ermitteln
                        maxByteCountForDeBlocks -= lengthDeBlock;
                    }
                }
            }

            if (isTelegramToSend) {

                // Telegramm versenden.
                debug.fine("Telegramm zum Versand: " + tlsData.toString());
                sendOsi7Telegramm(tlsData, de);
            }
        }
    }

    //~ GET METHODEN ==========================================================

    /**
     * Ermittelt die nächste zu verwendende Jobnummer im Wertebereich.
     *
     * @return Jobnummer im Wertebereich
     */
    private int getJobNumber(int min, int max)
    {
        // Jobnummer hochzählen und prüfen, ob im Bereich zwischen min und max
        if ((_jobNumber < min) || (_jobNumber >= max)) {
            _jobNumber = min;
        } else {
            _jobNumber++;
        }

        return _jobNumber;
    }

    /**
     * Liefert einen DeBlock zurück, der in das aktuelle Einzeltelegramm eingefügt werden kann (Übereinstimmung von
     * Osi7Adresse, Fg, ID) und der kürzer als die noch verfügbare Restlänge des Einzeltelegramms ist.
     *
     * @param deBlocksToSend          Die noch zu versendenden De-Blöcke.
     * @param osi7Adress              Adresse, an die der zu suchende De-Block gesendet werden muss.
     * @param fg                      Funktionsgruppe, die der zu suchende De-Block haben werden muss.
     * @param id                      ID, , die der zu suchende De-Block haben werden muss.
     * @param maxByteCountForDeBlocks Maimale Länge, , die der zu suchende De-Block haben darf.
     *
     * @return Der gefundene DeBlock oder null, wenn keiner gefunden wurde.
     */
    private DeBlockToSend getNextDeBlockToSend(List<DeBlockToSend> deBlocksToSend, int osi7Adress, int fg, int id, int maxByteCountForDeBlocks) {
        DeBlockToSend deBlockToSend;
        DeBlock       deBlock;
        De            de;

        for (int i = 0; i < (deBlocksToSend.size()); i++) {
            deBlockToSend = deBlocksToSend.get(i);
            deBlock       = deBlockToSend.getDeBlock();
            de            = deBlockToSend.getDe();

            if (

            // Prüfen, ob nur DE-Blocke versendet werden sollen, bei denen
            // für die De eine De-Gutmeldung anliegt.
            (de.isOk() ||!_sendOnlyDeOk) && (deBlock.getOsi7Adress() == osi7Adress) && (deBlock.getFg() == fg) && (deBlock.getId() == id) && (deBlock.getTlsDeDataBlock().length <= maxByteCountForDeBlocks)) {
                return deBlocksToSend.remove(i);
            }
        }

        return null;  // kein passender Eintrag mehr vorhanden
    }

    /**
     * Ermittelt die Osi7Adresse für den De-Block. Dabei wird auch berücksichtigt, ob dieser De-Block gegebenfalls mit
     * Osi7Adresse = 0 gesendet werden muss. (Spezielle konfigurierbare Fg/Typ/Id Kombinationen, i.d.R. in der FG 254).
     *
     * @param deBlock De-Block, für den Osi7Adresse ermittelt werden soll.
     *
     * @return Die ermittelte Osi7Adresse.
     */
    private int getOsi7Adress(DeBlock deBlock) {
        int osi7Adress;

        // Prüfen, ob mit Osi7Adresse = 0 gesendet werden muss.
        if (_sendingOsi7AdressNull.isSendingOsi7AdressNull(deBlock)) {
            debug.fine("##############################################################" + debug.NEWLINE + "Telegramm wird mit Knotennummer 0 versendet" + ": FG=" + deBlock.getFg() + ", Typ=" + deBlock.getTyp() + ", Id=" + deBlock.getId() + debug.NEWLINE + "##############################################################");
            osi7Adress = 0;
        } else {
            osi7Adress = deBlock.getOsi7Adress();
        }

        return osi7Adress;
    }

    //~ METHODEN ==============================================================

    /**
     * When an object implementing interface <code>Runnable</code> is used to create a thread, starting the thread causes
     * the object's <code>run</code> method to be called in that separately executing thread.
     * <p>
     * The general contract of the method <code>run</code> is that it may take any action whatsoever.
     *
     * @see Thread#run()
     */
    public void run() {
        while (true) {
            _deBlocksToSend = _deBlockQueue.popAll();
            debug.fine("############################################### Warte auf Objekte in Queue zum versenden...");

            for (int i = 0; i < _deBlocksToSend.size(); i++) {
                debug.fine("De-Block [" + i + "]: " + _deBlocksToSend.get(i));
            }

            debug.fine("############################################### ...Objekte in Queue zum versenden gelesen und verarbeitet.");

            // De-Blöcke Dateitransfer ausfiltern
            List<DeBlockToSend> filteredList = new ArrayList<DeBlockToSend>();

            for (DeBlockToSend blocki : _deBlocksToSend) {
                DeBlock deBlock = blocki.getDeBlock();
                De      de      = blocki.getDe();

                if (_deBlocksMultiBlockTransfer.isDeBlockMultiBlockTransfer(deBlock)) {
                    MultiBlockTransfer bt = _blockTransfers.get(de);

                    if (bt == null) {
                        bt = new MultiBlockTransfer();
                        _blockTransfers.put(de, bt);
                    }

                    bt.addDeBlockToSend(blocki);
                } else {
                    filteredList.add(blocki);
                }
            }

            // 'normale' DE-Blöcke senden
            erzeugeOsi7Telegramme(filteredList, false);

            // Blocktransfer senden
            Iterator<Entry<De, MultiBlockTransfer>> it = _blockTransfers.entrySet().iterator();

            while (it.hasNext()) {
                Entry<De, MultiBlockTransfer> e = it.next();

                if (e.getValue().isComplete()) {
                    debug.fine("Sende Multiblock-Transfer: " + e.getValue());
                    erzeugeOsi7Telegramme(e.getValue().getDeBlocksToSend(), true);
                    it.remove();
                }
            }
        }
    }

    /**
     * Versendet das übergebene Telegramm
     *
     * @param tlsData
     */
    private void sendOsi7Telegramm(TlsData tlsData, De de) {

        // Prüfen, ob nur Telegramme versendet werden sollen, bei denen
        // für das SM ein DEVICE_CONNECTED Status anliegt.
        if ((de.getEak().getSm().getState() == NetworkLayerListener.DEVICE_CONNECTED) ||!_sendOnlyConnectedSM) {
            try {
                int destination = de.getEak().getSm().getOsi7Adress();

                // Prüfen, ob Zieladresse umgeroutet werden muss.
                destination = _newRouting.getDestination(destination);
                _networkLayer.sendData(destination, tlsData.toBytes(), false);

                // Für die Jobnummernverwaltung sowohl versandt Telegramme als auch die einzelnen DE-Blöcke merken
                // ACHTUNG: Zuordnung von Jobnummer zu DeBlock dient AUSSCHNLIESSLICH der Umsetzung von
                // positiven Quittungen. In einem Telegramm (für eine Jobnummer) dürfen deshalb
                // nie mehr als zwei DeBlocke enthalten sein, wenn Zuordnung gelingen soll.
                Osi7Telegram     osi7Telegram   = new Osi7Telegram(destination, tlsData.toBytes(), false);
                SingleTelegram[] singleTelegram = osi7Telegram.getSingleTelegram();

                for (int i = 0; i < singleTelegram.length; ++i) {
                    int jobNumber = singleTelegram[i].getJobNumber();

                    TlsApplicationLayer.addSendSingleTelegram(jobNumber, singleTelegram[i]);

                    DeBlock[] deBlock = singleTelegram[i].getDeBlock();

                    for (int j = 0; j < deBlock.length; ++j) {
                        TlsApplicationLayer.addSendDeBlock(jobNumber, deBlock[j]);
                    }
                }
            }
            catch (DestinationUnreachableException e) {
                debug.error("Daten konnten nicht versendet werden: " + e);
            }
            catch (IllegalOsi7TelegramStructureException e) {
                debug.error("Telegrammaufbau des zu versendenden Telegramms fehlerhaft: \n" + e);
                debug.error("Daten konnten nicht versendet werden: " + e);
            }
        } else {
            debug.warning("Daten wurden nicht versand, da Status des Zielknotens ungleich CONNECTED!" + debug.NEWLINE + de.getEak().getSm().toString());
        }
    }

    //~ INNERE KLASSEN ========================================================

    /**
     * Klasse zur Repr&auml;sentation von zusammenh&auml;ngend zu &uuml;bertragenden
     * DE-Bl&ouml;cken (mit gleicher Jobnummer).
     *
     * @author BitCtrl Systems GmbH, Gieseler
     *
     */
    class MultiBlockTransfer {

        /**
         * Liste der zugeh&ouml;rigen Bl&ouml;cke.
         */
        private List<DeBlockToSend> deBlocks = new ArrayList<DeBlockToSend>();

        //~ METHODEN ==========================================================

        /**
         * F&uuml;gt einen neuen Block zur Liste hinzu.
         *
         * @param block zu sendender Block
         */
        public void addDeBlockToSend(DeBlockToSend block) {
            DeBlock deBlock     = block.getDeBlock();
            int     blockNummer = deBlock.getDeData().readUnsignedLowHighByteAsInt();

            deBlock.getDeData().rewindReadPosition();

            // Transfer zur&uumk;cksetzen, wenn Blocknummer == 1
            if (blockNummer <= 1) {
                deBlocks.clear();
            }

            deBlocks.add(block);
        }

        //~ GET METHODEN ======================================================

        /**
         * Gibt die Liste der zu sendenden Bl&ouml;cke zur&uuml;ck.
         *
         * @return Liste der zu sendenden Bl&ouml;cke
         */
        public List<DeBlockToSend> getDeBlocksToSend() {
            return deBlocks;
        }

        /**
         * Test, ob alle zugeh&ouml;rigen Bl&ouml;cke vorhanden sind.
         *
         * @return true, wenn alle Bl&ouml;cke vorhanden, sonst false
         */
        public boolean isComplete() {
            if (deBlocks.size() == 0) {
                return false;
            }

            DeBlockToSend firstBlock = deBlocks.get(0);
            DeBlock       deBlock    = firstBlock.getDeBlock();

            // Test auf Abbruch
            if (deBlocks.size() == 1) {
                int blockNumber = deBlock.getDeData().readUnsignedLowHighByteAsInt();  // Blocknummer

                deBlock.getDeData().rewindReadPosition();

                if (blockNumber == 0) {

                    // Abbruch
                    return true;
                }
            }

            // die Anzahl der eigentlichen Datenblöcke steht im ersten Block
            deBlock.getDeData().readUnsignedLowHighByteAsInt();  // Blocknummer

            int numberOfBlocks = deBlock.getDeData().readUnsignedLowHighByteAsInt();

            deBlock.getDeData().rewindReadPosition();

            return numberOfBlocks == deBlocks.size();
        }

        //~ METHODEN ==========================================================

        /**
         * Ausgabe des Transfers mit Einzelinformationen.
         *
         * @return String-Darstellung
         */
        public String toString() {
            StringBuffer sb      = new StringBuffer();
            final String NEWLINE = System.getProperty("line.separator");

            if (deBlocks.size() == 0) {
                return "";
            }

            DeBlockToSend firstBlock = deBlocks.get(0);
            DeBlock       deBlock    = firstBlock.getDeBlock();

            deBlock.getDeData().readUnsignedLowHighByteAsInt();  // Blocknummer

            int numberOfBlocks = deBlock.getDeData().readUnsignedLowHighByteAsInt();

            deBlock.getDeData().rewindReadPosition();
            sb.append("Knotennummer:" + deBlock.getOsi7Adress() + " (" + deBlock.getOsi7AdressLocCode() + ")" + " DE:" + deBlock.getDeChannel() + " FG:" + deBlock.getFg() + " Id:" + deBlock.getId() + " Typ:" + deBlock.getTyp());
            sb.append(NEWLINE);
            sb.append("Anzahl Blöcke: " + numberOfBlocks + 1);

            return (sb.toString());
        }
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 11.09.09
