/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt
 *
 * de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.EakHeuschBoesefeldt wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion.nw.vrz;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.kex.tls.osi7.*;
import de.bsvrz.kex.tls.osi7.conversion.EakDefault;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Umsetungsmodul für EAK, wird für die WanCom-Kopplung der ZDF basierten VRZ Leverkusen und VRZ Recklinghausen mit der
 * DaV basierten VRZ NRW verwendet.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class EakHeuschBoesefeldt extends EakDefault {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/** DebugLogger für Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Erzeugt eine neue Instanz von EakDefault
	 *
	 * @param daf DaV
	 * @param sm Sm
	 * @param eak Eak
	 */
	public EakHeuschBoesefeldt(ClientDavInterface daf, Sm sm, Eak eak) {
		super(daf, sm, eak);
		debug.finer("EakHeuschBoesefeldt Klasse wurde erfolgreich geladen und instanziert.");
	}

// ------------------------------------------------------------------------
// INTERFACE METHODEN
// ------------------------------------------------------------------------


// --------------------- Interface EakConversionInterface ---------------------

	/**
	 * Verarbeitet ein EinzelTelegramm. Enthält das Einzeltelegramm als ersten De-Block einen De-Block für die
	 * DeKanalNummer 255, so werden die Informationen dieses Blocks hier durch das Eak ausgewertet (Zeitinformationen und
	 * Intervalldaten etc.). Die für das Einzeltelegramm ermittelten allgemeinen Information stehen dann in den
	 * Umsetzungsmodulen für die De allen De's des Einzeltelegramms zur Verfügung.
	 *
	 * @param singleTelegram Das zu verarbeitende Einzeltelegramm {@link de.bsvrz.kex.tls.osi7.SingleTelegram}.
	 */
	public void processSingleTelegram(SingleTelegram singleTelegram) {
		super.processSingleTelegram(singleTelegram);

		if ((singleTelegram.getCountDeBlock() > 0) && (singleTelegram.getDeBlock()[0].getDeChannel() == 255)) {
			DeBlock deBlock = singleTelegram.getDeBlock()[0];

			debug.finest("Verabeitung durch den EAK: " + Debug.NEWLINE + deBlock.toString());

			switch (deBlock.getTyp()) {
				case 250:
				case 251:
				case 252:
				case 253:
					processTimeStampUnixTime(deBlock.getDeData().readNextTlsDataBlock(4));// UnixZeitstempel
					break;
			}
		}

		// Spezielle Anpassung für Koomunikation mit den VRZ von Heusch/Boesefeldt in NRW.
		// Prüft, ob Telegramme von der "falschen" VRZ gesendet werden. Wegen Überschneidung der Konfiguration
		// auf der VRZ Leverkusen und der VRZ Recklinghausen können z. T. einzelne Telegramme doppelt verschickt werden.
		// Dabei weisen die fälschlicherweise von der VRZ Leverkusen versandeten Telegramme von Daten der VRZ Recklinghausen
		// die falsche Absenderkennung auf (analog für Recklinghausen).
		// Wird ein "doppelter" (unzulässiger) DE-Block erkannt, wird dieser aus dem Einzeltelegramm entfernt.

		//		for (int i = 0; i < singleTelegram.getCountDeBlock(); i++) {
		//			DeBlock deBlock = singleTelegram.getDeBlock()[i];
		//			if (TlsApplicationLayer.isCorrectSender(singleTelegram.getSender(),
		//			                                        singleTelegram.getOsi7Adress(),
		//			                                        deBlock.getDeChannel(),
		//			                                        deBlock.getFg())) {
		//				debug.error("Unzulässiger Absender für DE-Block: " + Debug.NEWLINE + deBlock.toString());
		//				singleTelegram.deleteDeBlock(i);
		//			}
		//		}
	}

// ------------------------------------------------------------------------
// SONSTIGE METHODEN
// ------------------------------------------------------------------------

	/**
	 * Wertet den Zeitanteil eines DE-Blocks von Heusch/Boesefeldt im Format UNIX Zeit (Sekunden seit 1.1.1970, übergeben
	 * in den 4-Byte aus und stellt das Ergebnis über die Methode {@link #getTimeStamp} zur Verfügung.
	 *
	 * @param _tlsData Der DeBlock-Anteil mit der Zeitinformation.
	 */
	void processTimeStampUnixTime(TlsData _tlsData) {
		// UNIX-Zeit in Sekunden-------- -------------------------------------------------------------------------------
		long _UnixZeit_ = _tlsData.readUnsignedVeryHighExtraHighHighLowByteAsInt();
		final long offsetHBtoUTC = -3600;// Zeitdifferenz in Sekunden der HB-UNIX-Zeit zu UTC.

		setTimeStamp(((_UnixZeit_ + offsetHBtoUTC) * 1000));
	}
}
